@@ -247,8 +247,10 @@ externalSubset(void *ctx, const xmlChar *name,
 	/*
 	 * On the fly encoding conversion if needed
 	 */
-	enc = xmlDetectCharEncoding(ctxt->input->cur, 4);
-	xmlSwitchEncoding(ctxt, enc);
+	if (ctxt->input->length >= 4) {
+	    enc = xmlDetectCharEncoding(ctxt->input->cur, 4);
+	    xmlSwitchEncoding(ctxt, enc);
+	}
 
 	if (input->filename == NULL)
 	    input->filename = (char *) xmlStrdup(SystemID);
@@ -378,14 +380,26 @@ getEntity(void *ctx, const xmlChar *name)
 	((ctxt->validate) || (ctxt->replaceEntities)) &&
 	(ret->children == NULL) &&
 	(ret->etype == XML_EXTERNAL_GENERAL_PARSED_ENTITY)) {
+	int val;
+
 	/*
 	 * for validation purposes we really need to fetch and
 	 * parse the external entity
 	 */
 	xmlNodePtr children;
 
-        xmlParseCtxtExternalEntity(ctxt, ret->URI, ret->ExternalID, &children);
-	xmlAddChildList((xmlNodePtr) ret, children);
+        val = xmlParseCtxtExternalEntity(ctxt, ret->URI,
+		                         ret->ExternalID, &children);
+	if (val == 0) {
+	    xmlAddChildList((xmlNodePtr) ret, children);
+	} else {
+	    ctxt->sax->error(ctxt, 
+	     "Failure to process entity %s\n", name);
+	    ctxt->wellFormed = 0;
+	    ctxt->valid = 0;
+	    ctxt->validate = 0;
+	    return(NULL);
+	}
 	ret->owner = 1;
     }
     return(ret);
@@ -648,18 +662,6 @@ unparsedEntityDecl(void *ctx, const xmlChar *name,
     xmlGenericError(xmlGenericErrorContext,
 	    "SAX.unparsedEntityDecl(%s, %s, %s, %s)\n",
             name, publicId, systemId, notationName);
-#endif
-#if 0
-    Done in xmlValidateDtdFinal now.
-    if (ctxt->validate && ctxt->wellFormed && ctxt->myDoc) {
-	int ret;
-	ret = xmlValidateNotationUse(&ctxt->vctxt, ctxt->myDoc,
-	                                      notationName);
-	if (ret == 0) {
-	    ctxt->wellFormed = 0;
-	    ctxt->valid = 0;
-	}
-    }
 #endif
     if (ctxt->inSubset == 1) {
 	ent = xmlAddDocEntity(ctxt->myDoc, name,
@@ -860,27 +862,37 @@ my_attribute(void *ctx, const xmlChar *fullname, const xmlChar *value,
         (name[0] == 'x') && (name[1] == 'm') && (name[2] == 'l') &&
         (name[3] == 'n') && (name[4] == 's') && (name[5] == 0)) {
 	xmlNsPtr nsret;
+	xmlChar *val;
 
-	if (value[0] != 0) {
+        if (!ctxt->replaceEntities) {
+	    ctxt->depth++;
+	    val = xmlStringDecodeEntities(ctxt, value, XML_SUBSTITUTE_REF,
+		                          0,0,0);
+	    ctxt->depth--;
+	} else {
+	    val = (xmlChar *) value;
+	}
+
+	if (val[0] != 0) {
 	    xmlURIPtr uri;
 
-	    uri = xmlParseURI((const char *)value);
+	    uri = xmlParseURI((const char *)val);
 	    if (uri == NULL) {
 		if ((ctxt->sax != NULL) && (ctxt->sax->warning != NULL))
 		    ctxt->sax->warning(ctxt->userData, 
-			 "nmlns: %s not a valid URI\n", value);
+			 "nmlns: %s not a valid URI\n", val);
 	    } else {
 		if (uri->scheme == NULL) {
 		    if ((ctxt->sax != NULL) && (ctxt->sax->warning != NULL))
 			ctxt->sax->warning(ctxt->userData, 
-			     "xmlns: URI %s is not absolute\n", value);
+			     "xmlns: URI %s is not absolute\n", val);
 		}
 		xmlFreeURI(uri);
 	    }
 	}
 
 	/* a default namespace definition */
-	nsret = xmlNewNs(ctxt->node, value, NULL);
+	nsret = xmlNewNs(ctxt->node, val, NULL);
 
 	/*
 	 * Validate also for namespace decls, they are attributes from
@@ -889,27 +901,39 @@ my_attribute(void *ctx, const xmlChar *fullname, const xmlChar *value,
         if (nsret != NULL && ctxt->validate && ctxt->wellFormed &&
 	    ctxt->myDoc && ctxt->myDoc->intSubset)
 	    ctxt->valid &= xmlValidateOneNamespace(&ctxt->vctxt, ctxt->myDoc,
-					   ctxt->node, prefix, nsret, value);
+					   ctxt->node, prefix, nsret, val);
 	if (name != NULL) 
 	    xmlFree(name);
 	if (nval != NULL)
 	    xmlFree(nval);
+	if (val != value)
+	    xmlFree(val);
 	return;
     }
     if ((!ctxt->html) &&
 	(ns != NULL) && (ns[0] == 'x') && (ns[1] == 'm') && (ns[2] == 'l') &&
         (ns[3] == 'n') && (ns[4] == 's') && (ns[5] == 0)) {
 	xmlNsPtr nsret;
+	xmlChar *val;
 
-	if (value[0] == 0) {
+        if (!ctxt->replaceEntities) {
+	    ctxt->depth++;
+	    val = xmlStringDecodeEntities(ctxt, value, XML_SUBSTITUTE_REF,
+		                          0,0,0);
+	    ctxt->depth--;
+	} else {
+	    val = (xmlChar *) value;
+	}
+
+	if (val[0] == 0) {
 	    if ((ctxt->sax != NULL) && (ctxt->sax->error != NULL))
 		ctxt->sax->error(ctxt->userData, 
 		     "Empty namespace name for prefix %s\n", name);
 	}
-	if ((ctxt->pedantic != 0) && (value[0] != 0)) {
+	if ((ctxt->pedantic != 0) && (val[0] != 0)) {
 	    xmlURIPtr uri;
 
-	    uri = xmlParseURI((const char *)value);
+	    uri = xmlParseURI((const char *)val);
 	    if (uri == NULL) {
 		if ((ctxt->sax != NULL) && (ctxt->sax->warning != NULL))
 		    ctxt->sax->warning(ctxt->userData, 
@@ -925,7 +949,7 @@ my_attribute(void *ctx, const xmlChar *fullname, const xmlChar *value,
 	}
 
 	/* a standard namespace definition */
-	nsret = xmlNewNs(ctxt->node, value, name);
+	nsret = xmlNewNs(ctxt->node, val, name);
 	xmlFree(ns);
 	/*
 	 * Validate also for namespace decls, they are attributes from
@@ -939,6 +963,8 @@ my_attribute(void *ctx, const xmlChar *fullname, const xmlChar *value,
 	    xmlFree(name);
 	if (nval != NULL)
 	    xmlFree(nval);
+	if (val != value)
+	    xmlFree(val);
 	return;
     }
 
@@ -1032,8 +1058,9 @@ my_attribute(void *ctx, const xmlChar *fullname, const xmlChar *value,
 	    ctxt->valid &= xmlValidateOneAttribute(&ctxt->vctxt, ctxt->myDoc,
 					       ctxt->node, ret, value);
 	}
-    } else if (((ctxt->replaceEntities == 0) && (ctxt->external != 2)) ||
-	       ((ctxt->replaceEntities != 0) && (ctxt->inSubset == 0))) {
+    } else if (((ctxt->loadsubset & XML_SKIP_IDS) == 0) &&
+	       (((ctxt->replaceEntities == 0) && (ctxt->external != 2)) ||
+	        ((ctxt->replaceEntities != 0) && (ctxt->inSubset == 0)))) {
         /*
 	 * when validating, the ID registration is done at the attribute
 	 * validation level. Otherwise we have to do specific handling here.
