@@ -7,7 +7,7 @@
 #
 # daniel@veillard.com
 #
-import sys
+import os, sys
 import string
 import glob
 
@@ -182,9 +182,10 @@ class index:
 	      if self.functions.has_key(id):
 	          up = idx.functions[id]
 	          self.functions[id].update(None, up.type, up.info, up.extra)
-	      else:
-	          if idx.functions[id].static == 0:
-		      self.functions[id] = idx.functions[id]
+	 #     else:
+	 #         print "Function %s from %s is not declared in headers" % (
+	#	        id, idx.functions[id].module)
+	 # TODO: do the same for variables.
 
      def analyze_dict(self, type, dict):
          count = 0
@@ -225,6 +226,8 @@ ignored_words = {
   "LIBXML_DLL_IMPORT": (0, "Special macro to flag external keywords"),
   "__declspec": (3, "Windows keyword"),
   "ATTRIBUTE_UNUSED": (0, "macro keyword"),
+  "LIBEXSLT_PUBLIC": (0, "macro keyword"),
+  "X_IN_Y": (5, "macro function builder"),
 }
 
 class CLexer:
@@ -746,6 +749,7 @@ class CParser:
 		 name = token[1]
 		 signature = self.signature
 		 if signature != None:
+		     type = string.split(type, '(')[0]
 		     d = self.mergeFunctionComment(name,
 			     ((type, None), signature), 1)
 		     self.index.add(name, self.filename, not self.is_header,
@@ -854,7 +858,7 @@ class CParser:
 	 name = None
 	 self.comment = None
 	 comment = ""
-	 value = ""
+	 value = "0"
          while token != None:
 	     if token[0] == "sep" and token[1] == "{":
 	         token = self.token()
@@ -875,9 +879,9 @@ class CParser:
 			 self.enums.append((name, value, comment))
 		     name = token[1]
 		     comment = ""
-		     value = ""
 		     token = self.token()
 		     if token[0] == "op" and token[1][0] == "=":
+		         value = ""
 		         if len(token[1]) > 1:
 			     value = token[1][1:]
 		         token = self.token()
@@ -885,6 +889,12 @@ class CParser:
 			       token[1] != '}'):
 			     value = value + token[1]
 			     token = self.token()
+		     else:
+		         try:
+			     value = "%d" % (int(value) + 1)
+			 except:
+			     print "Failed to compute value of enum %s" % (name)
+			     value=""
 		     if token[0] == "sep" and token[1] == ",":
 			 token = self.token()
 	     else:
@@ -1206,7 +1216,7 @@ class CParser:
 			 self.index.add(self.name, self.filename, static,
 			                "function", d)
 			 token = self.token()
-		     if token[0] == "sep" and token[1] == "{":
+		     elif token[0] == "sep" and token[1] == "{":
 		         d = self.mergeFunctionComment(self.name,
 				 ((type, None), self.signature), static)
 			 self.index.add(self.name, self.filename, static,
@@ -1297,7 +1307,7 @@ class docBuilder:
 	 self.scanModules()
          
      def modulename_file(self, file):
-         module = string.split(file, '/')[-1]
+         module = os.path.basename(file)
 	 if module[-2:] == '.h':
 	     module = module[:-2]
 	 return module
@@ -1342,15 +1352,20 @@ class docBuilder:
 	     output.write("    <struct name='%s' file='%s' type='%s'" % (
 	              name, self.modulename_file(id.module), id.info))
 	     name = id.info[7:]
-	     if self.idx.structs.has_key(name):
+	     if self.idx.structs.has_key(name) and ( \
+	        type(self.idx.structs[name].info) == type(()) or
+		type(self.idx.structs[name].info) == type([])):
 	         output.write(">\n");
-		 for field in self.idx.structs[name].info:
-		     desc = field[2]
-		     if desc == None:
-		         desc = ''
-	             else:
-		         desc = escape(desc)
-		     output.write("      <field name='%s' type='%s' info='%s'/>\n" % (field[1] , field[0], desc))
+		 try:
+		     for field in self.idx.structs[name].info:
+			 desc = field[2]
+			 if desc == None:
+			     desc = ''
+			 else:
+			     desc = escape(desc)
+			 output.write("      <field name='%s' type='%s' info='%s'/>\n" % (field[1] , field[0], desc))
+		 except:
+		     print "Failed to serialize struct %s" % (name)
 		 output.write("    </struct>\n")
 	     else:
 	         output.write("/>\n");
@@ -1358,6 +1373,15 @@ class docBuilder:
 	     output.write("    <typedef name='%s' file='%s' type='%s'/>\n" % (
 	              name, self.modulename_file(id.module), id.info))
 
+     def serialize_variable(self, output, name):
+         id = self.idx.variables[name]
+	 if id.info != None:
+	     output.write("    <variable name='%s' file='%s' type='%s'/>\n" % (
+		     name, self.modulename_file(id.module), id.info))
+	 else:
+	     output.write("    <variable name='%s' file='%s'/>\n" % (
+	             name, self.modulename_file(id.module)))
+	              
      def serialize_function(self, output, name):
          id = self.idx.functions[name]
          output.write("    <%s name='%s' file='%s'>\n" % (id.type, name,
@@ -1419,6 +1443,10 @@ class docBuilder:
 	 typedefs.sort()
 	 for typedef in typedefs:
 	     self.serialize_typedef(output, typedef)
+	 variables = self.idx.variables.keys()
+	 variables.sort()
+	 for variable in variables:
+	     self.serialize_variable(output, variable)
 	 functions = self.idx.functions.keys()
 	 functions.sort()
 	 for function in functions:
@@ -1433,17 +1461,23 @@ def rebuild():
     if glob.glob("../parser.c") != [] :
         print "Rebuilding API description for libxml2"
 	builder = docBuilder("libxml2", ["..", "../include/libxml"],
-	                     ["xmlwin32version.h", "tst.c"])
+	                     ["xmlwin32version.h", "tst.c",
+			      "schemasInternals.h", "xmlschemas" ])
     elif glob.glob("../libxslt/transform.c") != [] :
         print "Rebuilding API description for libxslt"
 	builder = docBuilder("libxslt", ["../libxslt"],
-	                     ["win32config.h", "tst.c"])
+	                     ["win32config.h", "libxslt.h", "tst.c"])
     else:
         print "rebuild() failed, unable to guess the module"
 	return None
     builder.scan()
     builder.analyze()
     builder.serialize()
+    if glob.glob("../libexslt/exslt.c") != [] :
+        extra = docBuilder("libexslt", ["../libexslt"], ["libexslt.h"])
+	extra.scan()
+	extra.analyze()
+	extra.serialize()
     return builder
 
 #
