@@ -76,6 +76,7 @@
 #include <libxml/parser.h> /* for xmlStr(n)casecmp() */
 #include <libxml/nanohttp.h>
 #include <libxml/globals.h>
+#include <libxml/uri.h>
 
 /**
  * A couple portability macros
@@ -1411,9 +1412,11 @@ xmlNanoHTTPAuthHeader(void *ctx) {
 }
 
 /**
- * xmlNanoHTTPContentLength
+ * xmlNanoHTTPContentLength:
  * @ctx:  the HTTP context
  *
+ * Provides the specified content length from the HTTP header.
+ *
  * Return the specified content length from the HTTP header.  Note that
  * a value of -1 indicates that the content length element was not included in
  * the response header.
@@ -1426,11 +1429,13 @@ xmlNanoHTTPContentLength( void * ctx ) {
 }
 
 /**
- * xmlNanoHTTPFetchContent
+ * xmlNanoHTTPFetchContent:
  * @ctx:  the HTTP context
  * @ptr:  pointer to set to the content buffer.
  * @len:  integer pointer to hold the length of the content
  *
+ * Check if all the content was read
+ *
  * Returns 0 if all the content was read and available, returns
  * -1 if received content length was less than specified or an error 
  * occurred.
