@@ -593,13 +593,15 @@ static void processNode(xmlTextReaderPtr reader) {
     name = xmlTextReaderName(reader);
     if (name == NULL)
 	name = xmlStrdup(BAD_CAST "--");
+
     value = xmlTextReaderValue(reader);
 
-    printf("%d %d %s %d", 
+    printf("%d %d %s %d %d", 
 	    xmlTextReaderDepth(reader),
 	    xmlTextReaderNodeType(reader),
 	    name,
-	    xmlTextReaderIsEmptyElement(reader));
+	    xmlTextReaderIsEmptyElement(reader),
+	    xmlTextReaderHasValue(reader));
     xmlFree(name);
     if (value == NULL)
 	printf("\n");
@@ -1066,6 +1068,10 @@ static void parseAndPrintFile(char *filename) {
 	xmlRelaxNGValidCtxtPtr ctxt;
 	int ret;
 
+	if ((timing) && (!repeat)) {
+	    startTimer();
+	}
+
 	ctxt = xmlRelaxNGNewValidCtxt(relaxngschemas);
 	xmlRelaxNGSetValidErrors(ctxt,
 		(xmlRelaxNGValidityErrorFunc) fprintf,
@@ -1081,6 +1087,9 @@ static void parseAndPrintFile(char *filename) {
 		   filename);
 	}
 	xmlRelaxNGFreeValidCtxt(ctxt);
+	if ((timing) && (!repeat)) {
+	    endTimer("Validating");
+	}
 #endif
     }
 
@@ -1240,7 +1249,8 @@ main(int argc, char **argv) {
     int i, acount;
     int files = 0;
     int version = 0;
-
+    const char* indent;
+    
     if (argc <= 1) {
 	usage(argv[0]);
 	return(1);
@@ -1445,6 +1455,12 @@ main(int argc, char **argv) {
 	xmlRegisterNodeDefault(registerNode);
 	xmlDeregisterNodeDefault(deregisterNode);
     }
+    
+    indent = getenv("XMLLINT_INDENT");
+    if(indent != NULL) {
+	xmlTreeIndentString = indent;
+    }
+    
 
     xmlLineNumbersDefault(1);
     if (loaddtd != 0)
@@ -1470,6 +1486,9 @@ main(int argc, char **argv) {
     if (relaxng != NULL) {
 	xmlRelaxNGParserCtxtPtr ctxt;
 
+	if (timing) {
+	    startTimer();
+	}
 	ctxt = xmlRelaxNGNewParserCtxt(relaxng);
 	xmlRelaxNGSetParserErrors(ctxt,
 		(xmlRelaxNGValidityErrorFunc) fprintf,
@@ -1477,6 +1496,9 @@ main(int argc, char **argv) {
 		stderr);
 	relaxngschemas = xmlRelaxNGParse(ctxt);
 	xmlRelaxNGFreeParserCtxt(ctxt);
+	if (timing) {
+	    endTimer("Compiling the schemas");
+	}
     }
 #endif
     for (i = 1; i < argc ; i++) {
